import json
import os
import subprocess
import sys
from math import ceil
from pathlib import Path

import bioblend.galaxy
import lib
from bioblend.galaxy import dataset_collections
from ruamel.yaml import YAML

# Where we will look for our configuration file.
PROFILE_SEARCH_PATH = ['~/.abm/profile.yml', '.abm-profile.yml']

# Deprecated. Do not use.
datasets = {
    "dna": [
        "ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR013/ERR013101/ERR013101_1.fastq.gz",
        "ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR015/ERR015526/ERR015526_2.fastq.gz",
        "ftp://ftp.sra.ebi.ac.uk/vol1/fastq/SRR592/SRR592109/SRR592109_2.fastq.gz",
        "ftp://ftp.sra.ebi.ac.uk/vol1/fastq/SRR233/SRR233167/SRR233167_2.fastq.gz",
        "ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR047/ERR047678/ERR047678_1.fastq.gz",
        "ftp://ftp-trace.ncbi.nlm.nih.gov/ReferenceSamples/giab/data/AshkenazimTrio/HG003_NA24149_father/NIST_Illumina_2x250bps/reads/D2_S1_L002_R1_002.fastq.gz",
    ],
    "rna": [],
}


def try_for(f, limit=3):
    """
    Tries to invoke the function f. If the function f fails it will be retried
    *limit* number of times.

    :param f: the function to invoke
    :param limit: how many times the function will be retried
    :return: the result of calling f()
    """
    count = 0
    running = True
    result = None
    while running:
        try:
            count += 1
            result = f()
            running = False
        except Exception as e:
            if count >= limit:
                raise e
    return result


class Context:
    """
    The context object that contains information to connect to a Galaxy instance.

    GALAXY_SERVER: the URL of the Galaxy server to connect to
    API_KEY      : a user's API key to make API calls on the Galaxy instance
    KUBECONFIG:  : the kubeconfig file needed to make changes via Helm
    """

    def __init__(self, *args):
        if len(args) == 1:
            arg = args[0]
            if type(arg) == str:
                (
                    self.GALAXY_SERVER,
                    self.API_KEY,
                    self.KUBECONFIG,
                    self.MASTER_KEY,
                ) = parse_profile(arg)
            elif type(arg) == dict:
                self.GALAXY_SERVER = arg['GALAXY_SERVER']
                self.API_KEY = arg['API_KEY']
                self.KUBECONFIG = arg['KUBECONFIG']
                if 'MASTER_KEY' in arg:
                    self.MASTER_KEY = arg['MASTER_KEY']
                else:
                    self.MASTER_KEY = None
            else:
                raise Exception(f'Invalid arg for Context: {type(arg)}')
        elif len(args) == 3 or len(args) == 4:
            self.GALAXY_SERVER = args[0]
            self.API_KEY = args[1]
            self.KUBECONFIG = args[2]
            if len(args) == 4:
                self.MASTER_KEY = args[3]
        else:
            raise Exception(
                f'Invalid args for Context. Expected one or four, found {len(args)}'
            )


def print_json(obj, indent=2):
    print(json.dumps(obj, indent=indent))


def print_yaml(obj):
    get_yaml_parser().dump(obj, sys.stdout)


def connect(context: Context, use_master_key=False):
    """
    Create a connection to the Galaxy instance

    :return: a GalaxyInstance object
    """
    if context.GALAXY_SERVER is None:
        print('ERROR: The Galaxy server URL has not been set.  Please check your')
        print('       configuration in ~/.abm/profile.yml and try again.')
        sys.exit(1)
    if context.API_KEY is None:
        print('ERROR: The Galaxy API key has not been set.  Please check your')
        print('       configuration in ~/.abm/profile.yml and try again.')
        sys.exit(1)
    key = context.API_KEY
    if use_master_key:
        if context.MASTER_KEY is None:
            print('ERROR: The Galaxy master key has not been set.  Please check your')
            print('       configuration in ~/.abm/profile.yml and try again.')
            sys.exit(1)
        key = context.MASTER_KEY
    gi = bioblend.galaxy.GalaxyInstance(url=context.GALAXY_SERVER, key=key)
    gi.max_get_attempts = 3
    gi.get_retry_delay = 1
    return gi


def _set_active_profile(profile_name: str):
    """
    Unused.

    :param profile_name:
    :return:
    """
    lib.GALAXY_SERVER, lib.API_KEY, lib.KUBECONFIG, lib.MASTER_KEY = parse_profile(
        profile_name
    )
    return lib.GALAXY_SERVER != None


def get_context(profile_name: str):
    return Context(profile_name)


def get_yaml_parser():
    """
    Returns a singleton instance of a YAML parser.

    :return: a YAML parser.
    """
    if lib.parser is None:
        lib.parser = YAML()
    return lib.parser


def load_profiles():
    '''
    Load the profile configuration file.

    :return: a dictionary containing the YAML content of the configuration.
    '''
    yaml = get_yaml_parser()
    profiles = {}
    for profile_path in PROFILE_SEARCH_PATH:
        profile_path = os.path.expanduser(profile_path)
        if os.path.exists(profile_path):
            with open(profile_path, 'r') as f:
                # print(f'Loading profile from {profile_path}')
                profiles = yaml.load(f)
            break
    return profiles


def save_profiles(profiles: dict):
    """
    Write the ABM configuration file.

    :param profiles: the configuration to be saved.
    :return: None
    """
    yaml = get_yaml_parser()
    for profile_path in PROFILE_SEARCH_PATH:
        path = os.path.expanduser(profile_path)
        if os.path.exists(path):
            with open(path, 'w') as f:
                yaml.dump(profiles, f)
            print(f"Saved profiles to {path}")
            return


def parse_profile(profile_name: str):
    '''
    Parse the profile containing Galaxy URLs and API keys.

    :param profile_name: path to the profile to parse
    :return: a tuple containing the Galaxy URL, API key, and path to the kubeconfig
    '''
    nones = (None, None, None, None)
    profiles = load_profiles()
    if profiles is None:
        print(f'ERROR: Could not locate an abm profile file in {PROFILE_SEARCH_PATH}')
        return nones
    if profile_name not in profiles:
        print(f'ERROR: {profile_name} is not the name of a valid profile.')
        keys = list(profiles.keys())
        if len(keys) > 0:
            quoted_keys = (
                ', '.join([f"'{k}'" for k in keys[0:-2]]) + f", and '{keys[-1]}'"
            )
            print(f'The defined profile names are: {quoted_keys}')
        return nones
    profile = profiles[profile_name]
    kube = None
    master = 'galaxypassword'
    if 'kube' in profile:
        kube = os.path.expanduser(profile['kube'])
    if 'master' in profile:
        master = profile['master']
    return (profile['url'], profile['key'], kube, master)


def run(command, env: dict = None):
    """
    Runs a command on the local machine.  Used to invoke the helm and kubectl
    executables.

    :param command: the command to be invoked
    :param env: environment variables for the command.
    :return:
    """
    if env is None:
        env = os.environ
    result = subprocess.run(command.split(), capture_output=True, env=env)
    if result.returncode != 0:
        raise RuntimeError(result.stderr.decode('utf-8').strip())
    return result.stdout.decode('utf-8').strip()


def get_env(context: Context):
    """
    Creates a copy of the environment variables as returned by os.environ.
    :param context: Ignored
    :return: a dictionary of the environment variables
    """
    copy = os.environ.copy()
    for key, value in context.__dict__.items():
        if value is not None:
            copy[key] = value
    return copy


def find_executable(name):
    """
    Used the which command on the local machine to find the full path to an
    executable.

    :param name: the name of a command line executable or script.
    :return: the full path to the executable or an empty string if the executable is not found.
    """
    return run(f"which {name}")


# This is all the job metrics returned by Galaxy
#    "cpuacct.usage",
#     "end_epoch",
#     "galaxy_memory_mb",
#     "galaxy_slots",
#     "memory.failcnt",
#     "memory.limit_in_bytes",
#     "memory.max_usage_in_bytes",
#     "memory.memsw.limit_in_bytes",
#     "memory.memsw.max_usage_in_bytes",
#     "memory.oom_control.oom_kill_disable",
#     "memory.oom_control.under_oom",
#     "memory.soft_limit_in_bytes",
#     "memtotal",
#     "processor_count",
#     "runtime_seconds",
#     "start_epoch",
#     "swaptotal",
#     "uname"

# Columns to be defined when generating CSV files.
table_header = [
    "id",
    "history_id",
    "history_name",
    "state",
    "tool_id",
    "invocation_id",
    "workflow_id",
    "cpuacct.usage",
    # "end_epoch",
    "galaxy_memory_mb",
    "galaxy_slots",
    # "memory.failcnt",
    "memory.limit_in_bytes",
    "memory.peak",
    # "memory.max_usage_in_bytes",
    # "memory.memsw.limit_in_bytes",
    # "memory.memsw.max_usage_in_bytes",
    # "memory.oom_control.oom_kill_disable",
    # "memory.oom_control.under_oom",
    "memory.soft_limit_in_bytes",
    "memtotal",
    "processor_count",
    "runtime_seconds",
    # "start_epoch",
    # "swaptotal",
    # "uname"
]


def print_table_header():
    """
    Prints the table header suitable for inclusion in CSV files.

    :return: None. The table header is printed to stdout.
    """
    print(','.join(table_header))


history_name_cache = dict()


def get_history_name(gi, hid: str) -> str:
    if hid in history_name_cache:
        return history_name_cache[hid]
    history = gi.histories.show_history(hid)
    if history is None:
        return 'unknown'
    name = history['name']
    history_name_cache[hid] = name
    return name


def summarize_metrics(gi, jobs: list):
    table = []
    # table.append(header)
    # print(','.join(header))
    for job in jobs:
        job_metrics = gi.jobs.get_metrics(job['id'])
        row = []
        toolid = job.get('tool_id', 'unknown')
        if '/' in toolid:
            parts = toolid.split('/')
            toolid = f'{parts[-2]}/{parts[-1]}'
        metrics = metrics_to_dict(job_metrics, table_header)
        metrics['id'] = job.get('id', 'unknown')
        hid = job.get('history_id', 'unknown')
        metrics['history_id'] = hid
        metrics['history_name'] = get_history_name(gi, hid)
        metrics['state'] = job.get('state', 'unknown')
        metrics['tool_id'] = toolid
        metrics['invocation_id'] = job.get('invocation_id', 'unknown')
        for key in table_header:
            if key in metrics:
                row.append(metrics[key])
            else:
                row.append('')
        # print(','.join(row), end='\n')
        table.append(row)
    return table


def print_markdown_table(table: list) -> None:
    print('| Tool ID | History | State | Memory (GB) | Runtime (sec)|')
    print('|---|---|---:|---:|---:|')
    GB = 1024 * 1024 * 1024
    for row in table[1:]:
        # memory = ''
        # if row[11] != '':
        #     memory = float(row[11]) / GB
        #     if memory < 0.1:
        #         memory = 0.1
        #     memory = f"{memory:3.1f}"
        history = row[2]
        state = row[3]
        tool_id = row[4]
        # cpu = '' if row[7] == '' else float(row[7]) / 10**9
        memory = '' if row[11] == '' else f"{max(0.1, float(row[11]) / GB):3.1f}"
        runtime = '' if row[15] == '' else f"{max(1, float(row[15])):5.0f}"
        print(f'| {tool_id} | {history} | {state} | {memory} | {runtime} |')


def metrics_to_dict(metrics: list, accept: list):
    result = dict()
    for m in metrics:
        key = m['name']
        if key in accept:
            result[key] = m['raw_value']
    return result


def get_keys(d: dict):
    result = []
    for key in d.keys():
        result.append(key)
    result.sort()
    return result


def find_history(gi, name_or_id):
    history = None
    try:
        history = gi.histories.show_history(name_or_id)
    except:
        pass

    if history is not None:
        return history['id']
    history = gi.histories.get_histories(name=name_or_id)
    if history is None:
        return name_or_id
    if len(history) == 0:
        return None
    return history[0]['id']


def find_dataset(gi, history_id, name_or_id):
    try:
        dataset = gi.datasets.show_dataset(name=name_or_id)
        return dataset['id']
    except:
        pass

    try:
        dataset = gi.datasets.show_dataset(name_or_id)
        return dataset['id']
    except:
        pass
    return None
    # print("Calling get_datasets")
    # datasets = gi.datasets.get_datasets(history_id=history_id, name=name_or_id)
    # if datasets is None:
    #     print("Not found")
    #     return None
    # if len(datasets) == 0:
    #     print("No datasets found (len == 0)")
    #     return None
    # return datasets[0]['id']


def find_config(name: str) -> str:
    if os.path.exists(".abm"):
        if os.path.exists(f".abm/{name}"):
            return f".abm/{name}"
    config = os.path.join(Path.home(), ".abm", name)
    if os.path.exists(config):
        return config
    return None


def _get_dataset_data(gi, name_or_id):
    print(f"Getting dataset data for {name_or_id}")

    def make_result(data):
        return {
            'id': data['id'],
            'size': data['file_size'],
            'history': data['history_id'],
        }

    try:
        ds = gi.datasets.show_dataset(name_or_id)
        print(f"Got dataset data for {name_or_id} {ds['id']}")
        return make_result(ds)
    except Exception as e:
        print(f"Failed to get dataset data for {name_or_id}")
        pass

    try:
        print("Getting all datasets")
        datasets = gi.datasets.get_datasets(
            name=name_or_id
        )  # , deleted=True, purged=True)
        print(f"List of datasets for {name_or_id} is {len(datasets)}")
        for ds in datasets:
            # print_json(ds)
            state = True
            if 'state' in ds:
                state = ds['state'] == 'ok'
            if state and not ds['deleted'] and ds['visible']:
                # The dict returned by get_datasets does not include the input
                # file sizes so we need to make another call to show_datasets
                print(f"Getting dataset data for {ds['id']}")
                return make_result(gi.datasets.show_dataset(ds['id']))
            else:
                print(f"Skipping dataset {ds['id']}")
                print_json(ds)
    except Exception as e:
        pass

    return None


def _make_dataset_element(name, value):
    # print(f"Making dataset element for {name} = {value}({type(value)})")
    return dataset_collections.HistoryDatasetElement(name=name, id=value)


def get_float_key(column: int):
    def get_key(row: list):
        if row[column] == '':
            return -1
        return float(row[column])

    return get_key


def get_str_key(column: int):
    # print(f"Getting string key for column {column}")
    def get_key(row: list):
        # print(f"Sorting by column {column} key {row[column]}")
        return row[column]

    return get_key



import json
from pprint import pprint

from common import Context, find_executable, get_env, run


def pods(context: Context, args: list):
    namespace = '-n galaxy'
    if len(args) > 0:
        if args[0] == 'all':
            namespace = '-A'
        else:
            namespace = f"-n {args[0]}"

    kubectl = find_executable('kubectl')
    if kubectl is None:
        print("ERROR: kubectl is not on the $PATH")
        return

    print(run(f"{kubectl} get pods {namespace}"))


def url(context: Context, args: list):
    print(get_url(context, args))


def get_url(context: Context, args: list):
    # pprint(context.__dict__)
    # return
    kubectl = find_executable('kubectl')
    if kubectl is None:
        print("ERROR: kubectl is not on the $PATH")
        return

    namespace = 'galaxy'
    name = 'galaxy'
    if len(args) > 0:
        namespace = args[0]
    if len(args) > 1:
        name = args[1]
    command = f"{kubectl} get svc -n {namespace} {name}-nginx -o json"
    result = run(command, get_env(context))
    data = json.loads(result)
    ports = data['spec']['ports'][0]
    protocol = ports['name']
    port = ports['port']
    ip = data['status']['loadBalancer']['ingress'][0]['ip']
    return f"{protocol}://{ip}:{port}/{name}/"



import json
import os
import sys

sys.path.append(os.path.dirname(os.path.realpath(__file__)))

# Where the workflow invocation data returned by Galaxy will be saved.
INVOCATIONS_DIR = "invocations"
#  Where workflow runtime metrics will be saved.
METRICS_DIR = "metrics"

# Global instance of a YAML parser so we can reuse it if needed.
parser = None


# Keys used in various dictionaries.
class Keys:
    NAME = 'name'
    RUNS = 'runs'
    INPUTS = 'inputs'
    REFERENCE_DATA = 'reference_data'
    WORKFLOW_ID = 'workflow_id'
    DATASET_ID = 'dataset_id'
    COLLECTION = 'collection'
    HISTORY_BASE_NAME = 'output_history_base_name'
    HISTORY_NAME = 'history_name'


# def get_master_api_key():
#     '''
#     Get the master API key from the environment or configuration file.
#     '''
#     if 'GALAXY_MASTER_API_KEY' in os.environ:
#         return os.environ['GALAXY_MASTER_API_KEY']
#     config_path = os.path.expanduser("~/.abm/config.yml")
#     if not os.path.exists(config_path):
#         raise RuntimeError(f"ERROR: Configuration file not found: {config_path}")
#     with open(config_path, 'r') as f:
#         config = yaml.safe_load(f)
#     key = config.get('GALAXY_MASTER_API_KEY', None)
#     if key == None:
#         raise RuntimeError("ERROR: GALAXY_MASTER_API_KEY not found in config.yml")
#     return key



import os
from pathlib import Path

import yaml
from common import (Context, get_yaml_parser, load_profiles, print_json,
                    print_yaml, save_profiles)


def do_list(context: Context, args: list):
    profiles = load_profiles()
    print(f"Loaded {len(profiles)} profiles")
    for profile in profiles:
        print(f"{profile}\t{profiles[profile]['url']}")


def create(context: Context, args: list):
    if len(args) != 2:
        print("USAGE: abm config create <cloud> /path/to/kube.config")
        return
    profile_name = args[0]
    profiles = load_profiles()
    if profile_name in profiles:
        print("ERROR: a cloud configuration with that name already exists.")
        return
    profile = {'url': "", 'key': '', 'kube': args[1]}
    profiles[profile_name] = profile
    save_profiles(profiles)
    print_json(profile)


def remove(context: Context, args: list):
    if len(args) == 0:
        print("USAGE: abm config remove <cloud> [<cloud>...]")
        return
    profiles = load_profiles()
    for profile_name in args:
        if profile_name in profiles:
            del profiles[profile_name]
        else:
            print("ERROR: now cloud configuration with that name.")
    save_profiles(profiles)
    print_yaml(profiles)


def key(context: Context, args: list):
    if len(args) != 2:
        print(f"USAGE: abm config key <cloud> <key>")
        return
    profile_name = args[0]
    key = args[1]
    profiles = load_profiles()
    if not profile_name in profiles:
        print(f"ERROR: Unknown cloud {profile_name}")
        return
    profile = profiles[profile_name]
    profile['key'] = key
    save_profiles(profiles)
    print_json(profile)


def url(context: Context, args: list):
    if len(args) != 2:
        print(f"USAGE: abm config url <cloud> <url>")
        return
    profile_name = args[0]
    url = args[1]
    profiles = load_profiles()
    if not profile_name in profiles:
        print(f"ERROR: Unknown cloud {profile_name}")
        return
    profile = profiles[profile_name]
    profile['url'] = url
    save_profiles(profiles)
    print_json(profile)


def show(context: Context, args: list):
    if len(args) != 1:
        print("USAGE: abm config show <cloud>")
        return
    profiles = load_profiles()
    if args[0] not in profiles:
        print(f"ERROR: No such cloud {args[0]}")
        return
    print_json(profiles[args[0]])


def workflows(context: Context, args: list):
    userfile = os.path.join(Path.home(), ".abm", "workflows.yml")
    if len(args) == 0 or args[0] in ['list', 'ls']:
        workflows = _load_config(userfile)
        if workflows is None:
            return
        print(f"Workflows defined in {userfile}")
        for key, url in workflows.items():
            print(f"{key:10} {url}")
    elif args[0] in ['delete', 'del', 'rm']:
        print(
            f"Deleting workflows is not supported at this time. Please edit {userfile} directly."
        )
    elif args[0] in ['add', 'new']:
        print(
            f"Adding workflows is not supported at this time. Please edit {userfile} directly."
        )
    else:
        print(f"ERROR: Unrecognized command {args[0]}")


def datasets(context: Context, args: list):
    userfile = os.path.join(Path.home(), ".abm", "datasets.yml")
    if len(args) == 0 or args[0] in ['list', 'ls']:
        datasets = _load_config(userfile)
        if datasets is None:
            return
        print(f"Datasets defined in {userfile}")
        for key, url in datasets.items():
            print(f"{key:10} {url}")
    elif args[0] in ['delete', 'del', 'rm']:
        print(
            f"Deleting datasets is not supported at this time. Please edit {userfile} directly."
        )
    elif args[0] in ['add', 'new']:
        print(
            f"Adding datasets is not supported at this time. Please edit {userfile} directly."
        )
    else:
        print(f"ERROR: Unrecognized command {args[0]}")


def histories(context: Context, args: list):
    userfile = os.path.join(Path.home(), ".abm", "histories.yml")
    if len(args) == 0 or args[0] in ['list', 'ls']:
        histories = _load_config(userfile)
        if histories is None:
            return
        print(f"Datasets defined in {userfile}")
        for key, url in histories.items():
            print(f"{key:10} {url}")
    elif args[0] in ['delete', 'del', 'rm']:
        print(
            f"Deleting history entries is not supported at this time. Please edit {userfile} directly."
        )
    elif args[0] in ['add', 'new']:
        print(
            f"Adding dataset entries is not supported at this time. Please edit {userfile} directly."
        )
    else:
        print(f"ERROR: Unrecognized command {args[0]}")


# def _load_dataset_config(configfile):
#     if not os.path.exists(configfile):
#         print("ERROR: this instance has not been configured to import datasets.")
#         print(f"Please configure {configfile} to enable dataset imports.")
#         return None
#     return _load_config(configfile)


# def _load_workflow_config(userfile):
#     if not os.path.exists(userfile):
#         print("ERROR: this instance has not been configured to import workflows.")
#         print(f"Please configure {userfile} to enable workflow imports.")
#         return None
#     return _load_config(userfile)


def _load_config(filepath):
    if not os.path.exists(filepath):
        print(f"ERROR: configuration file not found: {filepath}")
        return None
    with open(filepath, 'r') as f:
        return yaml.safe_load(f)



import argparse
import json
import os
import time

from common import Context, find_executable, get_env, run


def rollback(context: Context, args: list):
    helm = find_executable('helm')
    # helm = 'helm'
    if helm is None:
        print('ERROR: helm is not available on the $PATH')
        return

    print(
        f"Rolling back deployment on {context.GALAXY_SERVER} KUBECONFIG: {context.KUBECONFIG}"
    )
    if len(args) > 0:
        command = f"{helm} rollback " + ' '.join(args)
    else:
        command = f"{helm} rollback galaxy -n galaxy"
    run(command, get_env(context))


def update(context: Context, args: list):
    if len(args) == 0:
        print(f'USAGE: abm <cloud> helm update <values> <namespace> <chart>')
        return
    values = args[0]
    namespace = 'galaxy' if args[1] is None else args[1]
    chart = 'galaxy/galaxy' if args[2] is None else args[2]

    helm = find_executable('helm')
    if helm is None:
        print('ERROR: helm is not available on the $PATH')
        return False

    if not os.path.exists(values):
        print(f"ERROR: Rules file not found: {values}")
        return False

    print(f"Applying rules {values} to {context.GALAXY_SERVER}")
    print(f"Helm update namespace: {namespace}")
    print(f"Helm update chart: {chart}")
    # command = f'{helm} upgrade galaxy {chart} -n {namespace} --reuse-values --set-file jobs.rules."container_mapper_rules\.yml".content={rules}'
    command = f'{helm} upgrade galaxy {chart} -n {namespace} --reuse-values -f {values}'
    env = get_env(context)
    try:
        result = run(command, env)
    except RuntimeError as e:
        print(f"Unable to helm upgrade {context.GALAXY_SERVER}")
        print(e)
        return False

    if result is None:
        return False

    print('Waiting for the new deployments to come online')
    # Give kubernetes a moment to start processing the update.
    time.sleep(5)
    wait_until_ready(namespace, env)
    return True


def update_cli(context: Context, args: list):
    """
    Runs the ``helm upgrade`` command on the cluster to update the
    *jobs.rules.container_mapper_rules* configuration.

    :param args:  a list containing the path to the rules YAML file.  See the
      *running* directory for examples.
    :return:
    """
    if context.KUBECONFIG is None:
        print("ERROR: No kubeconfig is specified in the profile")
        return False

    if len(args) == 0:
        print("ERROR: No rules specified.")
        return False

    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--values')
    parser.add_argument('-n', '--namespace', default='galaxy')
    parser.add_argument('chart', default='galaxy/galaxy')

    params = parser.parse_args(args)
    update(context, [params.values, params.namespace, params.chart])


def wait(context: Context, args: list):
    namespace = args[0] if len(args) > 0 else 'galaxy'
    wait_until_ready(namespace)  # , get_env(context))


def filter(lines: list, item: str):
    result = []
    for line in lines:
        if item in line:
            result.append(line)
    return result


def wait_for(kubectl: str, namespace: str, name: str, env: dict):
    print(f"Waiting for {name} on {env['GALAXY_SERVER']} to be in the Running state")
    waiting = True
    while waiting:
        # TODO The namespace should be parameterized
        result = run(f"{kubectl} get pods -n {namespace}", env)
        if result is None:
            break
        lines = result.split('\n')
        pods = filter(lines, name)
        if len(pods) == 0:
            print(f"ERROR: there are no pods named {name} in namespace {namespace}")
        if len(pods) == 1:
            tokens = pods[0].split()
            waiting = tokens[2] != 'Running'
        if waiting:
            print(f'Pods: {len(pods)}')
            for pod in pods:
                print(pod)
            time.sleep(30)
    print(f"{name} is running")


# def wait_until_ready(namespace: str, env: dict):
#     kubectl = find_executable('kubectl')
#     if kubectl is None:
#         print('ERROR: kubectl is not available on the $PATH')
#         return
#     wait_for(kubectl, namespace, 'galaxy-job', env)
#     wait_for(kubectl, namespace, 'galaxy-web', env)
#     wait_for(kubectl, namespace, 'galaxy-workflow', env)
def wait_until_ready(namespace: str, env: dict):
    kubectl = find_executable('kubectl')
    data = run(f"{kubectl} get deployment -n {namespace} -o json", env)
    deployment_data = json.loads(data)
    deployments = list()
    for deployment in deployment_data['items']:
        metadata = deployment['metadata']
        name = metadata['name']
        if 'job' in name or 'web' in name or 'workflow' in name:
            deployments.append(name)
    for deployment in deployments:
        print(
            run(
                f"{kubectl} rollout status deployment -n {namespace} {deployment} --watch",
                env,
            )
        )


def _list(context: Context, args: list):
    print("Not implemented")



#!/usr/bin/env python3

"""
The Automated Benchmarking Tool

Copyright 2024 The Galaxy Project. All rights reserved.

"""

import logging
import os
import sys
from pprint import pprint

import yaml
# These imports are required because we need Python to be load them to the
# symbol table so the parse_menu method can find them in globals()
from lib import (benchmark, cloudlaunch, config, dataset, experiment, folder,
                 helm, history, invocation, job, kubectl, library, users,
                 workflow)
from lib.common import Context

from abm import getVersion

log = logging.getLogger('abm')
handler = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')

log.setLevel(logging.ERROR)
handler.setLevel(logging.ERROR)
handler.setFormatter(formatter)
log.addHandler(handler)


# VERSION = '2.0.0-dev'

BOLD = '\033[1m'
CLEAR = '\033[0m'


def bold(text: str):
    """
    Wraps the text in ANSI control sequences to generate bold text in the terminal.

    :param text: the text to be made bold
    :return: the original string wrapped in ANSI control sequences
    """
    return f"{BOLD}{text}{CLEAR}"


help_args = ['help', '-h', '--help']
version_args = ['-v', '--version', 'version']

# Commands that do not depend on a cloud instance
stand_alone_commands = []


def head(text):
    print(bold(text))


def command_list(commands: list):
    return '|'.join(bold(c) for c in commands)


def copyright():
    print(f"    Copyright 2024 The Galaxy Project. All Rights Reserved.\n")


def print_main_help(menu_data):
    print()
    head("    DESCRIPTION")
    print("        Workflow and data management for remote Galaxy instances")
    print()
    head("    SYNOPSIS")
    print(f"        abm COMMAND [SUBCOMMAND] [OPTIONS]")
    print()
    head("    COMMANDS")
    for menu_item in menu_data:
        print(f"        {command_list(menu_item['name'])}")
        print(f"            {menu_item['help']}")
    print(f"        {command_list(['version', '-v', '--version'])}")
    print("            print the program version and exit")
    print(f"        {command_list(help_args)}")
    print("            print this help screen and exit")
    print()
    head("    NOTES")
    print(
        f"        Available SUBCOMMANDS and OPTIONS depend on the command. Use the {bold('help')} subcommand"
    )
    print(f"        to learn more about each of the commands. For example:\n")
    print(f"        $> abm workflow help\n")
    copyright()


def print_help(menu_data, command):
    submenu = None
    for menu_item in menu_data:
        if command in menu_item['name']:
            submenu = menu_item
            break
    if submenu is None:
        # print_main_help(menu_data)
        print(f"No help for {command} is available")
        return

    print()
    head("    SYNOPSIS")
    if submenu['name'][0] in stand_alone_commands:
        print(f"        abm {command} SUBCOMMAND <ARGS>\n")
    else:
        print(f"        abm [cloud] {command} SUBCOMMAND <ARGS>\n")
    head("    DESCRIPTION")
    print(f"        {submenu['help']}\n")
    head("    SUBCOMMANDS")
    for menu_item in submenu['menu']:
        print(
            f"        {'|'.join(bold(x) for x in menu_item['name'])} {menu_item['params'] if 'params' in menu_item else ''}"
        )
        print(f"            {menu_item['help']}")
    print(f"        {bold('help')}")
    print("            print this help screen and exit")
    print()
    copyright()


all_commands = {}


def get_menu(name: str):
    if name in all_commands:
        return all_commands[name]
    menu = dict()
    all_commands[name] = menu
    return menu


def register_handler(name: str, commands: list, handler):
    log.debug("Registering handler for menu: %s", name)
    menu = get_menu(name)
    for command in commands:
        menu[command] = handler


def alias(shortcut, fullname):
    all_commands[shortcut] = all_commands[fullname]


def parse_menu():
    log.debug('parse_menu')
    menu_config = f'{os.path.dirname(os.path.abspath(__file__))}/lib/menu.yml'
    if not os.path.exists(menu_config):
        log.error(f"ERROR: Unable to load the menu configuration from {menu_config}")
        # TODO Throw an exception that can be caught at the appropriate level
        sys.exit(1)
    with open(menu_config) as f:
        menu_data = yaml.safe_load(f)
    for main_menu_item in menu_data:
        # Use the first name in the list as the main name for the item. The
        # others will be aliased below.
        name = main_menu_item['name'][0]
        if main_menu_item.get('standalone', False):
            for item_name in main_menu_item['name']:
                stand_alone_commands.append(item_name)
        log.debug('Menu name: %s', name)
        for submenu_item in main_menu_item['menu']:
            handler = globals()
            handler_name = submenu_item['handler']
            log.debug('Submenu item: %s', handler_name)
            for part in handler_name.split('.'):
                # log.debug("Part: %s", part)
                if type(handler) is not dict:
                    handler = handler.__dict__
                if not part in handler:
                    print(f"ERROR: {part} not found")
                else:
                    handler = handler[part]
            if isinstance(handler, dict):
                log.error(f"Handler not found {handler_name}")
                # TODO Throw and excpetion that can be caught at the appropriate level.
                sys.exit(1)
            register_handler(name, submenu_item['name'], handler)
        for command_alias in main_menu_item['name'][1:]:
            alias(command_alias, name)
    return menu_data


def version():
    version = getVersion()
    print()
    print(f"    Galaxy Automated Benchmarking v{version}")
    copyright()


def _get_logopt(args: list):
    OPTS = ['-log', '--log', '-logging', '--logging']
    for i in range(len(args)):
        if args[i] in OPTS:
            return i
    return -1


def entrypoint():
    # Check if log level is being set
    logopt = _get_logopt(sys.argv)
    if logopt >= 0:
        if logopt + 1 >= len(sys.argv):
            print("ERROR: no log level provided")
            return

        level = sys.argv[logopt + 1].upper()
        if level not in [
            'DEBUG',
            'INFO',
            'WARN',
            'WARNING',
            'ERROR',
            'FATAL',
            'CRITICAL',
        ]:
            print(f"ERROR: Invalid logging level {sys.argv[logopt + 1]}")
            return
        print(f"Setting the log level to {level}")
        log.setLevel(level)
        global handler
        handler.setLevel(level)
        del sys.argv[logopt]
        del sys.argv[logopt]

    menu_data = parse_menu()

    if len(sys.argv) < 2 or sys.argv[1] in help_args:
        print_main_help(menu_data)
        return

    program = sys.argv[0]
    profile = sys.argv[1]
    if profile in version_args:
        version()
        return

    if profile in stand_alone_commands:
        command = profile
        profile = None
        if len(sys.argv) < 3:
            print_help(menu_data, command)
            return
        subcommand = sys.argv[2]
        if len(sys.argv) > 3:
            params = sys.argv[3:]
        else:
            params = []
    else:
        if len(sys.argv) < 3:
            print(f"{bold('ERROR')} Invalid command {bold(profile)}")
            return
        command = sys.argv[2]
        if command in version_args:
            version()
            return
        subcommand = None
        if len(sys.argv) > 3:
            subcommand = sys.argv[3]
        if len(sys.argv) > 4:
            params = sys.argv[4:]
        else:
            params = []

    if command in help_args:
        print_help(menu_data, profile)
        return
    if subcommand and subcommand in help_args:
        print_help(menu_data, command)
        return

    context = None
    if profile is not None:
        context = Context(profile)
    if command in all_commands:
        subcommands = all_commands[command]
        if subcommand not in subcommands:
            print(f'ERROR: unrecognized subcommand "{subcommand}"')
            print(f'Type "{program} {command} help" for more help.')
            return
        handler = subcommands[subcommand]
        handler(context, params)
    else:
        print(f'\n{bold("ERROR:")} Unknown command {bold({command})}')
        print_main_help(menu_data)


if __name__ == '__main__':
    entrypoint()



import argparse

from common import (Context, connect, get_float_key, get_str_key, print_json,
                    print_markdown_table, print_table_header, print_yaml,
                    summarize_metrics)


def doList(context: Context, args: list):
    wid = None
    hid = None
    while len(args) > 0:
        arg = args.pop(0)
        if args in ['-h', '--history']:
            hid = args.pop(0)
        elif arg in ['-w', '--workflow']:
            wid = args.pop(0)
        else:
            print(f'Invalid parameter: {arg}')
            return
    gi = connect(context)
    invocations = gi.invocations.get_invocations(workflow_id=wid, history_id=hid)
    print('ID\tState\tWorkflow\tHistory')
    for invocation in invocations:
        id = invocation['id']
        state = invocation['state']
        workflow = invocation['workflow_id']
        history = invocation['history_id']
        print(f'{id}\t{state}\t{workflow}\t{history}')


def show(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: no invocation ID was provided")
        return
    gi = connect(context)
    invocation = gi.invocations.show_invocation(args[0])
    print_yaml(invocation)


def summarize(context: Context, args: list):
    parser = argparse.ArgumentParser()
    parser.add_argument('id', nargs=1)
    parser.add_argument('--markdown', action='store_true')
    parser.add_argument('-s', '--sort-by', choices=['runtime', 'memory', 'tool'])
    argv = parser.parse_args(args)
    gi = connect(context)
    id = argv.id[0]
    all_jobs = []
    jobs = gi.jobs.get_jobs(invocation_id=id)
    for job in jobs:
        job['invocation_id'] = id
        job['workflow_id'] = ''
        all_jobs.append(job)
    table = summarize_metrics(gi, all_jobs)
    if argv.sort_by:
        reverse = True
        get_key = None
        if argv.sort_by == 'runtime':
            get_key = get_float_key(15)
        elif argv.sort_by == 'memory':
            get_key = get_float_key(11)
        elif argv.sort_by == 'tool':
            get_key = get_str_key(4)
            reverse = False
        table.sort(key=get_key, reverse=reverse)
    if argv.markdown:
        print_markdown_table(table)
    else:
        print_table_header()
        for row in table:
            print(','.join(row))



import os
import sys
import json
import yaml
import bioblend.galaxy


job_index = {}

def load_config():
    path = os.path.expanduser("~/.abm/profile.yml")
    with open(path) as f:
        profiles = yaml.safe_load(f)
    return profiles


def record_job(data):
    cloud = data['cloud']
    if cloud not in job_index:
        job_index[cloud] = []
    job_index[cloud].append(data)


def process(filepath: str):
    with open(filepath) as f:
        job_data = json.load(f)
    job_id = os.path.basename(filepath).replace(".json", "")
    job_data['job_id'] = job_id
    record_job(job_data)
    # status = 'Yes' if 'job_metrics' in job_data['metrics'] else 'No'
    # print(filepath, job_data['cloud'], status)


def run(directories):
    for dir in directories:
        for file in os.listdir(dir):
            input_path = os.path.join(dir, file)
            if os.path.isfile(input_path) and input_path.endswith(".json"):
                process(input_path)

    profiles = load_config()
    for cloud,jobs in job_index.items():
        if cloud not in profiles:
            print(f"ERROR: cloud {cloud} has not been defined")
            continue
        profile = profiles[cloud]
        gi = bioblend.galaxy.GalaxyInstance(url=profile['url'], key=profile['key'])
        for job in jobs:
            id = job['metrics']['id']
            try:
                job['metrics']['job_metrics'] = gi.jobs.get_metrics(id)
                output_path = f"../resource-prediction/metrics/TestD-5-fixed/{id}.json"
                try:
                    with open(output_path, 'w') as f:
                        f.write(json.dumps(job, indent=4))
                    print(f"Wrote {output_path}")
                except Exception as e:
                    print(e)
            except:
                # print(f"Unable to get metrics for job {id} from {cloud}")
                pass

    # gi = bioblend.galaxy.GalaxyInstance()
    # data = gi.jobs.get_metrics('97af82a28609d22c')
    # print(json.dumps(data, indent=4))


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print("Please provide a directory name.")
    else:
        run(sys.argv[1:])


import os
import sys
import argparse


def accept(line: str, tool: str, filters: list):
    if 'ok' not in line:
        return False
    if tool not in line:
        return False
    for f in filters:
        if f not in line:
            return False
    return True


def run(file: str, tool: str, filter:list):
    if not os.path.exists(file):
        print("ERROR: file not found")
        return
    with open(file) as f:
        lines = f.read().splitlines(keepends=False)

    # print(f"Loaded {len(lines)} lines")
    # for line in lines:
    #     print(line)
    print("Filters are")
    for f in filter:
        print(f)
    data = [ line for line in lines if accept(line, tool, filter)]
    runtime = 0
    memory = 0
    count = 0
    for line in data:
        parts = line.split('|')
        if parts[3].strip() == 'ok':
            print(line)
            runtime += float(parts[-2])
            memory += float(parts[-3])
            count += 1
        else:
            print(f"Rejecting line {parts[3]}: {line}")
    if count == 0:
        print("ERROR: No lines matched the given query")
        return

    print(f"{memory/count:3.3f}GB {runtime/count:5.2f}")


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('file', nargs=1)
    parser.add_argument('-t', '--tool', help='Tool name to filter on')
    parser.add_argument('-f', '--filter', help='A string to also filter on', nargs='+')
    args = parser.parse_args(sys.argv[1:])
    run(args.file[0], args.tool, args.filter)



import argparse
import json
import os
from pathlib import Path
from pprint import pprint

import yaml
from bioblend.galaxy import dataset_collections
from common import (Context, _get_dataset_data, _make_dataset_element, connect,
                    find_config, find_dataset, find_history, print_json)


def do_list(context: Context, argv: list):
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--state', help='list datasets in this state')
    parser.add_argument('--history', help='show datasets in the given history')
    parser.add_argument('-t', '--tool', help='only show datasets generate by this tool')
    args = parser.parse_args(argv)
    kwargs = {'limit': 10000, 'offset': 0, 'deleted': False}
    gi = connect(context)
    if args.state:
        kwargs['state'] = args.state
    if args.history:
        hid = find_history(gi, args.history)
        if hid is None:
            print("ERROR: No such history")
            return
        kwargs['history_id'] = hid
    if args.tool:
        kwargs['tool_id'] = args.tool

    # if len(args) > 0:
    #     if args[0] in ['-s', '--state']:
    #         if len(args) != 2:
    #             print("ERROR: Invalid command.")
    #             return
    #         kwargs['state'] = args[1]
    #     else:
    #         print(f"ERROR: Invalid parameter: {args[0]}")
    #         return
    # datasets = gi.datasets.get_datasets(limit=10000, offset=0)  # , deleted=True, purged=True)
    datasets = gi.datasets.get_datasets(**kwargs)
    if len(datasets) == 0:
        print('No datasets found')
        return
    print(f'Found {len(datasets)} datasets')
    print('ID\tHistory\tType\tDeleted\tState\tName')
    for dataset in datasets:
        state = dataset['state'] if 'state' in dataset else 'unknown'
        print(
            f"{dataset['id']}\t{dataset['history_id']}\t{dataset['history_content_type']}\t{dataset['deleted']}\t{state}\t{dataset['name']}"
        )


def clean(context: Context, args: list):
    if len(args) == 0:
        invalid_states = ['error', 'discarded', 'unknown']
    else:
        invalid_states = args
    gi = connect(context)
    datasets = gi.datasets.get_datasets(
        limit=10000, offset=0
    )  # , deleted=True, purged=True)
    if len(datasets) == 0:
        print('No datasets found')
        return
    for dataset in datasets:
        state = dataset['state'] if 'state' in dataset else 'unknown'
        if not dataset['deleted'] and state in invalid_states:
            gi.histories.delete_dataset(dataset['history_id'], dataset['id'], True)
            print(f"Removed {dataset['id']}\t{state}\t{dataset['name']}")


def show(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: no dataset ID provided")
        return
    gi = connect(context)
    result = gi.datasets.show_dataset(args[0])
    print(json.dumps(result, indent=4))


def get(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: no dataset ID provided")
        return
    gi = connect(context)
    result = gi.datasets.get_datasets(args[0])
    print(json.dumps(result, indent=4))


def delete(context: Context, args: list):
    # gi = connect(context)
    print("dataset delete not implemented")


def upload(context: Context, args: list):
    history = None
    name = None
    url = None
    gi = None
    while len(args) > 0:
        arg = args.pop(0)
        if arg in ['--hs', '--hist', '--history']:
            history = args.pop(0)
        elif arg in ['-c', '--create']:
            gi = connect(context)
            history = gi.histories.create_history(args.pop(0)).get('id')
        elif arg in ['-n', '--name']:
            name = args.pop(0)
        elif url is not None:
            print(f'ERROR: invalid option. URL already set to {arg}')
            return
        else:
            url = arg
    if history is None:
        print("ERROR: please specify or create a history.")
        return
    if gi is None:
        gi = connect(context)
    history = find_history(gi, history)
    if name:
        _import_from_url(gi, history, url, file_name=name)
    else:
        _import_from_url(gi, history, url)


def collection(context: Context, args: list):
    type = 'list:paired'
    collection_name = 'collection'
    elements = []
    hid = None
    gi = connect(context)
    while len(args) > 0:
        arg = args.pop(0)
        if arg == '-t' or arg == '--type':
            type = args.pop(0)
        elif arg == '-n' or arg == '--name':
            collection_name = args.pop(0)
        elif '=' in arg:
            name, value = arg.split('=')
            dataset = _get_dataset_data(gi, value)
            if dataset is None:
                print(f"ERROR: dataset not found {value}")
                return
            # print_json(dataset)
            if hid is None:
                hid = dataset['history']
            elif hid != dataset['history']:
                print('ERROR: Datasets must be in the same history')
            elements.append(_make_dataset_element(name, dataset['id']))

    if len(elements) == 0:
        print("ERROR: No dataset elements have been defined for the collection")
        return
    result = gi.histories.create_dataset_collection(
        history_id=hid,
        collection_description=dataset_collections.CollectionDescription(
            name=collection_name, type=type, elements=elements
        ),
    )
    print(json.dumps(result, indent=4))


def import_from_config(context: Context, args: list):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-c',
        '--create',
        help='create a new history for the dataset',
        required=False,
        default=None,
    )
    parser.add_argument(
        '-f',
        '--file',
        help='use instead of the datasets.yml',
        required=False,
        default=None,
    )
    parser.add_argument(
        '--history',
        help='add datasets to the given history',
        required=False,
        default=None,
    )
    parser.add_argument(
        '-n', '--name', help='set the name of the dataset', required=False, default=None
    )
    parser.add_argument('keys', help='the key of the dataset to import', nargs='+')
    gi = None
    history = None
    kwargs = {}
    argv = parser.parse_args(args)
    if argv.name is not None:
        if len(argv.keys) > 1:
            print("ERROR: cannot specify --name with multiple keys")
            return
        kwargs['file_name'] = argv.name

    if argv.create is not None and argv.history is not None:
        print("ERROR: cannot specify both --create and --history")
        return

    if argv.create is not None:
        gi = connect(context)
        history = gi.histories.create_history(argv.create).get('id')
    if argv.history is not None:
        gi = connect(context)
        history = find_history(gi, argv.history)
    if argv.file is not None:
        configfile = argv.file
        if not os.path.exists(configfile):
            print(f"ERROR: the specified file {configfile} was not found")
            return
    else:
        configfile = find_config("datasets.yml")
        if configfile is None:
            print("ERROR: ABM has not been configured to import datasets.")
            print(f"Please create {configfile}")
            return
    with open(configfile, 'r') as f:
        datasets = yaml.safe_load(f)
    if gi is None:
        gi = connect(context)
    for key in argv.keys:
        if not key in datasets:
            print(f"ERROR: dataset {key} has not been defined.")
        else:
            url = datasets[key]
            print(f"Importing {key} from {url}")
            response = gi.tools.put_url(url, history, **kwargs)
            print(json.dumps(response, indent=4))


def _import_from_url(gi, history, url, **kwargs):
    response = gi.tools.put_url(url, history, **kwargs)
    print(json.dumps(response, indent=4))


def download(context: Context, args: list):
    gi = connect(context)
    if len(args) == 0:
        print('ERROR: no dataset ID given')
        return
    elif len(args) > 1:
        pprint(gi.datasets.download_dataset(args[0], file_path=args[1]))
    else:
        pprint(gi.datasets.download_dataset(args[0]))


def find(context: Context, args: list):
    if len(args) == 0:
        print('ERROR: no dataset name given.')
        return
    gi = connect(context)
    datasets = gi.datasets.get_datasets(name=args[0])
    if len(datasets) == 0:
        print('WARNING: no datasets found with that name')
        return
    if len(datasets) > 1:
        print(f'WARNING: found {len(datasets)} datasets with that name.')

    for ds in datasets:
        pprint(ds)


def rename(context: Context, args: list):
    if len(args) != 3:
        print("ERROR: please provide the history ID, dataset ID, and new name.")
        return
    gi = connect(context)
    hid = find_history(gi, args[0])
    if hid is None:
        print("ERROR: no such history")
        return
    dsid = find_dataset(gi, hid, args[1])
    if dsid is None:
        print("ERROR: no such dataset")
        return
    response = gi.histories.update_dataset(hid, dsid, name=args[2])
    # result = {'state': response['state'], 'name': response['name']}
    print(json.dumps(response, indent=4))


def test(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: no dataset ID provided")
        return
    gi = connect(context)
    dataset = gi.datasets.show_dataset(args[0])
    if dataset is None:
        print(f"ERROR: no dataset with ID {args[0]}")
        return
    name = dataset['name']
    print(f"Found dataset with name {name}")
    print(yaml.dump(dataset))
    datasets = gi.datasets.get_datasets(name=name)
    if datasets is None or len(datasets) == 0:
        print(f"ERROR: could not find a dataset named {name}")
        return
    print(f"Found {len(datasets)} datasets")
    for dataset in datasets:
        print(f"{dataset['id']}\t{dataset['name']}")



import sys
import argparse

#
# parser = argparse.ArgumentParser()
# parser.add_argument('input_dirs', nargs='+')
# parser.add_argument('-t', '--tsv', action='store_true')
# parser.add_argument('-c', '--csv', action='store_true')
# parser.add_argument('--markdown', action='store_true')
# parser.add_argument('-s', '--sort-by', choices=['cpu', 'runtime', 'memory'])
#
# args = parser.parse_args(sys.argv[1:])
# print('tsv', args.tsv)
# print('csv', args.csv)
# print('md', args.markdown)
# print('sort', args.sort_by)
# print('type', type(args.input_dirs))
# for p in args.input_dirs:
#     print('path', p)
#

DATA = '''Name
ID
Type
Private DNS Zone Contributor
b12aa53e-6015-4669-85d0-8515ebb3ae7f
BuiltInRole
Start and Stop VM_
2aeb0baf-603a-49d6-b0f4-860c38e52692
CustomRole
Virtual Machine Administrator Login
1c0163c0-47e6-4577-8991-ea5c82e286e4
BuiltInRole
Virtual Machine Contributor
9980e02c-c2be-4d73-94e8-173b1dc7cf3c
BuiltInRole
Virtual Machine Data Access Administrator (preview)
66f75aeb-eabe-4b70-9f1e-c350c4c9ad04
BuiltInRole
Virtual Machine Local User Login
602da2ba-a5c2-41da-b01d-5360126ab525
BuiltInRole
Virtual Machine User Login
fb879df8-f326-4884-b1cf-06f3ad86be52
BuiltInRole
Windows 365 Network User
7eabc9a4-85f7-4f71-b8ab-75daaccc1033
BuiltInRole
Azure Arc Enabled Kubernetes Cluster User Role
00493d72-78f6-4148-b6c5-d3ce8e4799dd
BuiltInRole
Azure Kubernetes Service RBAC Writer
a7ffa36f-339b-4b5c-8bdf-e2c188b2c0eb
BuiltInRole
Azure Arc Kubernetes Admin
dffb1e0c-446f-4dde-a09f-99eb5cc68b96
BuiltInRole
Azure Arc Kubernetes Cluster Admin
8393591c-06b9-48a2-a542-1bd6b377f6a2
BuiltInRole
Azure Arc Kubernetes Viewer
63f0a09d-1495-4db4-a681-037d84835eb4
BuiltInRole
Azure Arc Kubernetes Writer
5b999177-9696-4545-85c7-50de3797e5a1
BuiltInRole
Azure Kubernetes Fleet Manager Contributor Role
63bb64ad-9799-4770-b5c3-24ed299a07bf
BuiltInRole
Azure Kubernetes Fleet Manager RBAC Admin
434fb43a-c01c-447e-9f67-c3ad923cfaba
BuiltInRole
Azure Kubernetes Fleet Manager RBAC Cluster Admin
18ab4d3d-a1bf-4477-8ad9-8359bc988f69
BuiltInRole
Azure Kubernetes Fleet Manager RBAC Reader
30b27cfc-9c84-438e-b0ce-70e35255df80
BuiltInRole
Azure Kubernetes Fleet Manager RBAC Writer
5af6afb3-c06c-4fa4-8848-71a8aee05683
BuiltInRole
Azure Kubernetes Service Cluster Admin Role
0ab0b1a8-8aac-4efd-b8c2-3ee1fb270be8
BuiltInRole
Azure Kubernetes Service Cluster Monitoring User
1afdec4b-e479-420e-99e7-f82237c7c5e6
BuiltInRole
Azure Kubernetes Service Cluster User Role
4abbcc35-e782-43d8-92c5-2d3f1bd2253f
BuiltInRole
Azure Kubernetes Service Contributor Role
ed7f3fbd-7b88-4dd4-9017-9adb7ce333f8
BuiltInRole
Azure Kubernetes Service Hybrid Cluster User Role
fc3f91a1-40bf-4439-8c46-45edbd83563a
BuiltInRole
Azure Kubernetes Service Hybrid Cluster Admin Role
b5092dac-c796-4349-8681-1a322a31c3f9
BuiltInRole
Azure Kubernetes Service Hybrid Contributor Role
e7037d40-443a-4434-a3fb-8cd202011e1d
BuiltInRole
Azure Kubernetes Service Policy Add-on Deployment
18ed5180-3e48-46fd-8541-4ea054d57064
BuiltInRole
Azure Kubernetes Service RBAC Admin
3498e952-d568-435e-9b2c-8d77e338d7f7
BuiltInRole
Azure Kubernetes Service RBAC Cluster Admin
b1ff04bb-8a4e-4dc4-8eb5-8693973ce19b
BuiltInRole
Azure Kubernetes Service RBAC Reader
7f6c6a51-bcf8-42ba-9220-52d62157d7db
BuiltInRole
Storage Blob Data Contributor
ba92f5b4-2d11-453d-a403-e96b0029c9fe
BuiltInRole
Storage Blob Data Owner
b7e6dc6d-f1e8-4753-8033-0f276bb0955b
BuiltInRole
Storage Blob Data Reader
2a2b9908-6ea1-4ae2-8e65-a410df84e7d1
BuiltInRole
Storage Blob Delegator
db58b8e5-c6ad-4a2a-8342-4190687cbf4a
BuiltInRole
'''

def run():
    count = 0
    role = ''
    for line in DATA.splitlines(keepends=False):
        # print("LINE:", line)
        count += 1
        if count % 3 == 1:
            # print(line, end='')
            role = line
        elif count % 3 == 0:
            print(f"{line} - {role}")


if __name__ == '__main__':
    run()


import argparse
import json
import logging
import os
import threading
import traceback
from datetime import timedelta
from pprint import pprint
from time import perf_counter

import benchmark
import helm
import yaml
from common import (Context, get_float_key, get_str_key, load_profiles,
                    print_markdown_table)

INVOCATIONS_DIR = "invocations"
METRICS_DIR = "metrics"

log = logging.getLogger('abm')


def run(context: Context, args: list):
    """
    Runs a single benchmark defined by *args[0]*

    :param args: a list that contains a single element, the path to a benchmark
      configuration file.

    :return: True if the benchmarks completed sucessfully. False otherwise.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('benchmark_path')
    parser.add_argument('-r', '--run-number', default=-1)
    argv = parser.parse_args(args)

    benchmark_path = argv.benchmark_path

    if not os.path.exists(benchmark_path):
        print(f"ERROR: Benchmarking configuration not found {benchmark_path}")
        return False

    with open(benchmark_path, 'r') as f:
        config = yaml.safe_load(f)
    config['start_at'] = argv.run_number
    print(f"Starting with run number {argv.run_number}")

    profiles = load_profiles()
    threads = []
    start = perf_counter()
    for cloud in config['cloud']:
        if cloud not in profiles:
            print(f"WARNING: No profile found for {cloud}")
            continue
        t = threading.Thread(target=run_on_cloud, args=(cloud, config))
        threads.append(t)
        print(f"Starting thread for {cloud}")
        t.start()
    print('Waiting for threads')
    for t in threads:
        t.join()

    end = perf_counter()
    print('All threads have terminated.')
    print(f"Execution time {timedelta(seconds=end - start)}")


def run_on_cloud(cloud: str, config: dict):
    print("------------------------")
    print(f"Benchmarking: {cloud}")
    context = Context(cloud)
    namespace = 'galaxy'
    chart = 'anvil/galaxykubeman'
    start = int(config['start_at'])
    print(f"Staring run number {start}")
    if start < 0:
        start = 1
    print(f"Staring run number {start}")
    end = start + config['runs']
    if 'galaxy' in config:
        namespace = config['galaxy']['namespace']
        chart = config['galaxy']['chart']
    if 'job_configs' in config and len(config['job_configs']) > 0:
        for conf in config['job_configs']:
            rules_file = f"rules/{conf}.yml"
            print(f"Applying {rules_file} namespace:{namespace} chart:{chart}")
            if not helm.update(context, [f"rules/{conf}.yml", namespace, chart]):
                log.warning(f"job configuration not found: rules/{conf}.yml")
                continue
            for workflow_conf in config['benchmark_confs']:
                for n in range(start, end):
                    history_name_prefix = f"{n} {cloud} {conf}"
                    benchmark.run(
                        context, workflow_conf, history_name_prefix, config['name']
                    )
    else:
        for workflow_conf in config['benchmark_confs']:
            for n in range(start, end):
                history_name_prefix = f"{n} {cloud}"
                benchmark.run(
                    context, workflow_conf, history_name_prefix, config['name']
                )


def test(context: Context, args: list):
    print(context.GALAXY_SERVER)
    if os.path.exists(args[0]):
        with open(args[0]) as f:
            data = yaml.safe_load(f)
        print(data)


def parse_toolid(id: str) -> str:
    parts = id.split('/')
    if len(parts) < 2:
        return f"{id},unknown"
    return f"{parts[-2]},{parts[-1]}"


def summarize(context: Context, args: list):
    """
    Parses all the files in the specified directory and prints metrics
    as CSV to stdout

    :param args[0]: The path to the directory containing metrics filees
    :return: None
    """
    markdown = False
    separator = None
    make_row = make_table_row
    header_row = "Run,Cloud,Job Conf,Workflow,History,Inputs,Tool,Tool Version,State,Slots,Memory,Runtime (Sec),CPU,Memory Limit (Bytes),Memory Max usage (Bytes)"

    parser = argparse.ArgumentParser()
    parser.add_argument('dirs', nargs='*')
    parser.add_argument('-c', '--csv', action='store_true')
    parser.add_argument('-t', '--tsv', action='store_true')
    parser.add_argument('-m', '--model', action='store_true')
    parser.add_argument('--markdown', action='store_true')
    parser.add_argument('-s', '--sort-by', choices=['runtime', 'memory', 'tool'])
    argv = parser.parse_args(args)

    count = 0
    if argv.csv:
        separator = ','
        count += 1
    if argv.tsv:
        separator = '\t'
        count += 1
    if argv.model:
        separator = ','
        make_row = make_model_row
        count += 1
    if argv.markdown:
        markdown = True
        count += 1

    if count == 0:
        print("ERROR: no output format selected")
        return
    if count > 1:
        print("ERROR: multiple output formats selected")
        return

    input_dirs = argv.dirs

    if len(input_dirs) == 0:
        input_dirs.append('metrics')

    if separator is None:
        separator = ','

    if markdown:
        print("|Run|Inputs|Job Conf|Tool|State|Runtime (Sec)|Max Memory (GB)|")
        print("|---|---|---|---|---|---:|---:|")
    else:
        print(header_row)

    table = list()
    GB = float(1073741824)
    for input_dir in input_dirs:
        for file in os.listdir(input_dir):
            input_path = os.path.join(input_dir, file)
            if not os.path.isfile(input_path) or not input_path.endswith('.json'):
                continue
            try:
                with open(input_path, 'r') as f:
                    data = json.load(f)
                if data['metrics']['tool_id'] == 'upload1':
                    # print('Ignoring upload tool')
                    continue
                row = make_row(data)
                table.append(row)
            except Exception as e:
                print(f"Unable to process {input_path}")
                print(e)
                traceback.print_exc()
                # Silently fail to allow the remainder of the table to be generated.
                # pass

    reverse = True
    if argv.sort_by:
        comp = get_str_key(6)
        if argv.sort_by == 'runtime':
            # key = 10
            comp = get_float_key(10)
        # elif argv.sort_by == 'cpu':
        #     comp = get_float_comparator(11)
        #     #key = 11
        elif argv.sort_by == 'memory':
            comp = get_float_key(13)
            # key = 13
        elif argv.sort_by == 'tool':
            # print('Getting string key accessor.')
            comp = get_str_key(6)
            reverse = False
        # table.sort(key=lambda row: -1 if row[key] == '' else float(row[key]), reverse=True)
        table.sort(key=comp, reverse=reverse)

    if markdown:
        for row in table:
            runtime = '' if len(row[10]) == 0 else f"{float(row[10]):4.1f}"
            # cpu = '' if len(row[11]) == 0 else f"{float(row[11])/10**9:4.1f}"
            memory = '' if len(row[13]) == 0 else f"{float(row[13])/GB:4.3f}"
            # memory = float(row[13]) / GB
            print(
                f"| {row[0]} | {row[5].split(' ')[0]} |{row[2]} | {row[6]} | {row[7]} | {runtime}  | {memory} |"
            )
    else:
        for row in table:
            print(separator.join([str(x) for x in row]))


accept_metrics = [
    'galaxy_slots',
    'galaxy_memory_mb',
    'runtime_seconds',
    'cpuacct.usage',
    'memory.limit_in_bytes',
    'memory.peak'
    #'memory.max_usage_in_bytes',
]  # ,'memory.soft_limit_in_bytes']


def make_table_row(data: dict):
    row = [
        str(data[key])
        for key in ['run', 'cloud', 'job_conf', 'workflow_id', 'history_id', 'inputs']
    ]
    row.append(parse_toolid(data['metrics']['tool_id']))
    row.append(data['metrics']['state'])
    for e in _get_metrics(data['metrics']['job_metrics']):
        row.append(e)
    return row


def make_model_row(data: dict):
    metrics = data['metrics']
    row = []
    row.append(metrics['id'])
    tool_id = metrics['tool_id']
    row.append(tool_id)
    row.append(tool_id.split('/')[-1])
    row.append(metrics['state'])
    job_metrics = parse_job_metrics(metrics['job_metrics'])
    row.append(job_metrics.get('memory.max_usage_in_bytes', 0))
    row.append(job_metrics.get('cpuacct.usage', 0))
    row.append(job_metrics.get('processor_count', 0))
    row.append(job_metrics.get('galaxy_slots', 0))
    row.append(job_metrics.get('runtime_seconds', 0))
    if len(data['ref_data_size']) == 0:
        row.append('0')
    else:
        row.append(data['ref_data_size'][0])
    for size in data['input_data_size']:
        row.append(size)
    return row


def _get_metrics(metrics: list):
    row = [''] * len(accept_metrics)
    for job_metrics in metrics:
        if job_metrics['name'] in accept_metrics:
            index = accept_metrics.index(job_metrics['name'])
            try:
                row[index] = job_metrics['raw_value']
            except:
                pass
    return row


def add_metrics_to_row(metrics_list: list, row: list):
    for job_metrics in metrics_list:
        if job_metrics['name'] in accept_metrics:
            index = accept_metrics.index(job_metrics['name'])
            try:
                row[index + 8] = job_metrics['raw_value']
            except:
                pass


def parse_job_metrics(metrics_list: list):
    metrics = {}
    for job_metrics in metrics_list:
        metrics[job_metrics['name']] = job_metrics['raw_value']
    return metrics



import os
import sys
import json
import yaml

workflow_path = '/Users/suderman/Workspaces/JHU/Experiments/variants/variants.yml'

if not os.path.exists(workflow_path):
    print("ERROR: workflow file not found")
else:
    with open(workflow_path) as f:
        benchmark = yaml.safe_load(f)

    for input in benchmark[0]['runs'][0]['inputs']:
        if 'paired' in input:
            for item in input['paired']:
                for key in item.keys():
                    print(key)
        elif 'dataset_id' in input:
            print(f"dataset ID: {input['dataset_id']}")
        elif 'value' in input:
            print(f"literal: {input['value']}")
        else:
            print("Invalid input definition")

